
#include "aenc.h"
#include <cassert>


AEnc::AEnc(const ACfg& acfg, std::string& outname)
	: prefix(acfg.prefix)
	, V(acfg.Vbits)
	, U(acfg.Ubits)
	, bitstream(outname)
	, valid(bitstream.isValid())
	, numEncoded(0)
{
	assert(U + V < 64);
	// init arithmetic coding engine
	{
		A = ((uint64_t)1 << U) - 1;
		B = 0;
		C = 0;
		A_temp = 0;
		B_temp = 0;

		if (prefix)
		{
			n = 2;
		}
		else
		{
			n = 1;
		}
	}
}


AEnc::~AEnc()
{
  assert( !valid );
}


void AEnc::encode( uint8_t index, const Pmf& pmf )
{
  assert( valid );

  // arithmetic encoding
  {
	  A_temp = A * pmf[index];
	  B_temp = B + (A * c_v(index, pmf));
	  z = find_z();
	  A = A_temp >> (V - z);
	  B = (B_temp & (((uint64_t)1 << (U + V - z)) - 1)) << z;
	  update_output();
  }

  // update symbol count
  numEncoded++;
}


void AEnc::terminate()
{
  assert( valid );
  // terminate arithmetic coding engine
  {
	  bool roundUp = false;

	  for (uint64_t i = 0; i < U + V - n; i++)
	  {
		  if (((B >> i) & 1) == 1)
		  {
			  roundUp = true;
			  break;
		  }
	  }

	  if (roundUp)
	  {
		  B = B + ((uint64_t)1 << (U + V - n));
		  if (B >= ((uint64_t)1 << (U + V)))
		  {
			  if (C >= 1)
			  {
				  pushBits(C, true);
				  if (C == 1)
				  {
					  C = 0;
				  }
				  else
				  {
					  C = 1;
				  }
			  }
			  B = B - ((uint64_t)1 << (U + V));
		  }
	  }
	  if (C >= 1)
	  {
		  pushBits(C, false);
		  C = 0;
	  }

	  for (uint64_t i = 1; i <= n; i++)
	  {
		  if (((B & ((uint64_t)1 << (U + V - i))) >> (U + V - i)) == 1)
		  {
			  pushBits(C, true);
		  }
		  else
		  {
			  pushBits(C, false);
		  }
	  }
  }

  // terminate bitstream
  bitstream.terminate( numEncoded );
  valid = false;
}

uint64_t AEnc::c_v(uint8_t index, const Pmf& pmf)
{
	uint64_t c_v_value = 0;
	for (uint64_t k = 0; k < index; k++)
	{
		c_v_value += pmf[k];
	}
	return c_v_value;
}

uint64_t AEnc::find_z()
{
	uint64_t zeros =0;
	for (uint64_t i = 0; i < U + V; i++)
	{
		if (((A_temp >> (U + V - (i + 1))) & 1) == 0)
		{
			zeros++;
		}
		else
		{
			break;
		}
	}
	return zeros;
}

void AEnc::pushBits(uint64_t c, bool isInverted )
{
	uint64_t invertBits = 0;

	if (isInverted)
	{
		invertBits = 1;
	}

	bitstream.addBit(0 + invertBits);
	if (1 + invertBits <= c)
	{
		for (uint64_t i = 0; i < c - (1 + invertBits); i++)
		{
			bitstream.addBit(1 - invertBits);
		}
	}
}

void AEnc::update_output()
{
	if (((B_temp >> (U + V)) & 1) == 1)
	{
		if (C > 1)
		{
			pushBits(C, true);
			C = 1;
		}
		else
		{
			pushBits(C, true);
			C = 0;
		}
	}

	for (uint64_t i = 1; i <= z; i++)
	{
		if (C >= 1)
		{
			if (((B_temp >> (U + V - i)) & 1) == 1)
			{
				C++;
			}
			else
			{
				pushBits(C, false);
				C = 1;
			}
		}
		else
		{
			if (((B_temp >> (U + V - i)) & 1) == 1)
			{
				pushBits(C, true);
			}
			else
			{
				C++;
			}
		}
	}
}
